# -*- coding: utf-8 -*-
# ----------------------------------------
# Copyright (C) 2010-2011  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import sys
import codecs
import csv
import mx.DateTime
from Tkinter import *
import tkFileDialog
import tkMessageBox
import tkSimpleDialog
if sys.version_info[0] >= 3:
    import tkinter.ttk as ttk
else:
    import ttk
from toureditorlib.tourfile import TourFile
from toureditorlib.tourgraph import TourGraph
from toureditorlib._version import __version__

# ----------------------------------------

_encodings = [
    'iso-8859-1',
    'iso-8859-2',
    'iso-8859-3',
    'iso-8859-4',
    'iso-8859-5',
    'iso-8859-6',
    'iso-8859-7',
    'iso-8859-8',
    'iso-8859-9',
    'iso-8859-10',
    'iso-8859-11',
    'iso-8859-12',
    'iso-8859-13',
    'iso-8859-14',
    'iso-8859-15',
    'iso-8859-16',
    'utf-8'
    ]

_filetypes = [
    ('CRP', '*.crp'),
    ('CSV', '*.csv'),
    ('GPX', '*.gpx'),
    ('all', '*')
    ]

# ----------------------------------------

class TourView (Toplevel):

    def __init__ (self, tourfile):

        Toplevel.__init__ (self)
        self.tourfile = tourfile
        self.listeners = []
        self.fixedRows = set()

        self.title('Tour Editor - '+tourfile.filename)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1)

        # 1st button frame
        bf = ttk.Frame (self)
        bf.grid (row=0, column=0, columnspan=2, sticky=W)

        l = ttk.Label (bf, text='Encoding:')
        l.pack (side=LEFT)

        self.encodingBox = ttk.Combobox(bf, values=_encodings)
        self.encodingBox.set (tourfile.encoding or 'iso-8859-1')
        self.encodingBox.pack (side=LEFT, fill=Y, expand=True)

        self.buttons = {}
        self.buttons['open'] = b = ttk.Button (bf, text='Open', command=self.open)
        b.pack(side=LEFT)

        self.buttons['save'] = b = ttk.Button (bf, text='Save', command=self.save, state=DISABLED)
        b.pack(side=LEFT)

        self.buttons['saveAs'] = b = ttk.Button (bf, text='Save As', command=self.saveAs)
        b.pack(side=LEFT)

        self.buttons['reload'] = b = ttk.Button (bf, text='Reload', command=self.reload, state=DISABLED)
        b.pack(side=LEFT)

        self.buttons['close'] = b = ttk.Button (bf, text='Close', command=self.close)
        b.pack(side=LEFT)

        # Treeview with scrollbars
        t = ttk.Treeview (self, columns=tourfile.columns)
        t.grid(row=1,column=0, sticky=NSEW)
        self.treeview = t
        t.column ('#0', width=60, stretch=False)
        for i in range(len(tourfile.columns)-1):
            t.column (i, width=100, stretch=False)
        t.tag_configure ('standstill', background='yellow')
        t.tag_configure ('fixed', foreground='red')

        self.vscr = ttk.Scrollbar (self, command=self.treeview.yview)
        t.config (yscrollcommand=self.vscr.set)
        self.vscr.grid(row=1,column=1, sticky=NS)

        self.hscr = ttk.Scrollbar (self, orient=HORIZONTAL, command=self.treeview.xview)
        t.config (xscrollcommand=self.hscr.set)
        self.hscr.grid(row=2,column=0, sticky=EW)

        t.bind ('<Double-Button-1>', self.doubleclick)
        self.bind ('<Control-a>', self.selectAll)
        self.bind ('<Control-n>', self.selectNone)

        # 2nd button frame
        bf = ttk.Frame (self)
        bf.grid (row=3, column=0, columnspan=2)

        self.buttons['meta'] = b = ttk.Button (bf, text='Metadata', command=self.metadata)
        b.pack(side=LEFT)

        self.buttons['hOffset'] = b = ttk.Button (bf, text='Height Offset', command=self.heightOffset)
        b.pack(side=LEFT)

        self.buttons['dOffset'] = b = ttk.Button (bf, text='Distance Offset', command=self.distanceOffset)
        b.pack(side=LEFT)

        self.buttons['delete'] = b = ttk.Button (bf, text='Delete Rows', command=self.delete)
        b.pack(side=LEFT)

        self.buttons['reverse'] = b = ttk.Button (bf, text='Reverse File', command=self.reverse)
        b.pack(side=LEFT)

        self.buttons['fix'] = b = ttk.Button (bf, text='Fix rows', command=self.fix)
        b.pack(side=LEFT)

        self.buttons['smooth'] = b = ttk.Button (bf, text='Smooth', command=self.smooth)
        b.pack(side=LEFT)

        self.buttons['graph'] = b = ttk.Button (bf, text='Show Graph', command=self.graph)
        b.pack(side=LEFT)

        self.geometry ('800x600')
        self.refresh()

    def registerListener (self, listener):
        self.listeners.append (listener)

    def deregisterListener (self, listener):
        self.listeners = [l for l in self.listeners if l != listener]

    def dirty (self, isdirty):
        if isdirty:
            self.buttons['save'].config (state=NORMAL)
            self.buttons['reload'].config (state=NORMAL)
        else:
            self.buttons['save'].config (state=DISABLED)
            self.buttons['reload'].config (state=DISABLED)

    def refresh (self):
        t = self.treeview
        items = t.get_children()
        if len(items) > 0:
            t.delete (*items)
        tf = self.tourfile
        for i in range(len(tf.columns)):
            t.heading (i, text=tf.columns[i])
        for i in range(len(tf)):
            point = tf.get(i)
            point = [point[0].strftime ('%H:%M:%S')] + point[1:]
            for j in range (1, len(point)):
               if point[j] == None: point[j] = ''
            tags = []
            if i > 0 and abs(tf.distance (i-1, i)) < .001:
                tags.append ('standstill')
            if i in self.fixedRows:
                tags.append ('fixed')
            t.insert ('', 'end', text=str(i+1), value=point, tags=tuple(tags))
        for l in self.listeners:
            l.refresh()

    def open(self):
        filename = tkFileDialog.askopenfilename (
            title='Open file', filetypes=_filetypes)
        if filename != None and filename != '':
            encoding = self.encodingBox.get()
            TourView (TourFile.read (filename, encoding))

    def save(self):
        if tkMessageBox.askokcancel (title='Save Tour File', message='Really overwrite contents of '+self.csv.filename+'?'):
            tf = self.tourfile
            tf.write (tf.filename, self.encodingBox.get(), tf.filetype)
            self.dirty (False)

    def saveAs(self):
        filename = tkFileDialog.asksaveasfilename (
            title='Save File', filetypes=_filetypes)
        if filename != None and filename != '':
            self.tourfile.write (filename, self.encodingBox.get())
            self.title (filename)
            self.dirty (False)

    def reload(self):
        tf = self.tourfile
        self.tourfile = TourFile.read (tf.filename, tf.encoding, tf.filetype)
        self.refresh()
        self.dirty (False)

    def close(self):
        for l in self.listeners:
            l.close()
        self.destroy()

    def delete (self):
        t = self.treeview
        tf = self.tourfile
        rows = [t.index(item) for item in t.selection()]
        rows.sort(reverse=True)
        for row in rows:
            tf.delete (row)
        self.refresh()
        self.dirty (True)

    def fix (self):
        t = self.treeview
        self.fixedRows = self.fixedRows.symmetric_difference(
            [t.index(item) for item in t.selection()])
        self.refresh()

    def doubleclick (self, event):
        t = self.treeview
        item = t.identify_row (event.y)
        col = t.identify_column (event.x)
        if item != '' and col != '#0':
            s = tkSimpleDialog.askstring (title='Enter value', prompt=t.heading(col, option='text'),
                                          initialvalue=t.set(item, col))
            t.set (item, col, s)
            self.tourfile.set (t.index(item), **{t.column(col, 'id'): s})
            self.dirty (True)
            for l in self.listeners:
                l.refresh()

    def selectAll (self, event):
        t = self.treeview
        t.selection_set (t.get_children())

    def selectNone (self, event):
        self.treeview.selection_set(())

    def metadata (self):
        metadataDialog = MetadataDialog(self)

    def heightOffset (self):
        t = self.treeview
        tf = self.tourfile
        offset = tkSimpleDialog.askfloat(title='Height Offset', prompt='Select offset [m] to add/subtract from height')
        if offset != None:
            for item in t.selection():
                value = int(t.set (item, 'height'))
                value += offset
                t.set (item, 'height', str(value))
                tf.put (t.index(item), height=value)
            self.dirty (True)

    def distanceOffset (self):
        t = self.treeview
        tf = self.tourfile
        offset = tkSimpleDialog.askfloat(title='Distance Offset', prompt='Select offset [km] to add/subtract from distance')
        if offset != None:
            offset = round(offset, 2)
            for item in t.selection():
                value = float(t.set (item, 'distance'))
                t.set (item, 'distance', str(value+offset))
                tf.put (t.index(item), distance=value+offset)
            self.dirty (True)

    def reverse (self):
        self.tourfile.reverse()
        self.refresh()
        self.dirty (True)

    def smooth (self):
        self.tourfile.smooth(self.fixedRows)
        self.refresh()
        self.dirty (True)

    def graph(self):
        tf = self.tourfile
        TourGraph (self)

# ----------------------------------------

class MetadataDialog (tkSimpleDialog.Dialog):

    def body(self, master):
        self.title('Metadata')

        tf = self.parent.tourfile

        ttk.Label (master, text='Title:').grid (row=0, column=0)
        self.tourtitle = ttk.Entry (master, width=20)
        if tf.title != None:
            self.tourtitle.insert ('0', tf.title)
        self.tourtitle.grid (row=0, column=1)

        ttk.Label (master, text='Date:').grid (row=1, column=0)
        self.date = ttk.Entry (master, width=10)
        if tf.date != None:
            self.date.insert ('0', tf.date.strftime("%Y-%m-%d"))
        self.date.grid (row=1, column=1)

        ttk.Label (master, text='Copyright').grid (row=2, column=0)
        self.copyright = ttk.Entry (master, width=20)
        if tf.copyright != None:
            self.copyright.insert ('0', tf.copyright)
        self.copyright.grid (row=2, column=1)

        ttk.Label (master, text='Length [km]').grid (row=3, column=0)
        ttk.Label (master, text=tf.length()).grid (row=3, column=1)

        ttk.Label (master, text='Total height difference [m]').grid (row=4, column=0)
        ttk.Label (master, text=tf.hdiff(False)).grid (row=4, column=1)

        ttk.Label (master, text='Net height difference [m]').grid (row=5, column=0)
        ttk.Label (master, text=tf.hdiff(True)).grid (row=5, column=1)

        ttk.Label (master, text='Average slope [%]').grid (row=6, column=0)
        ttk.Label (master, text=round(tf.slope(), 2)).grid (row=6, column=1)

        return None

    def apply(self):
        tf = self.parent.tourfile
        tf.title = self.tourtitle.get()
        date = self.date.get()
        if date != None and date != '':
            tf.date = mx.DateTime.strptime (date, "%Y-%m-%d")
        tf.copyright = self.copyright.get()
        self.result = 1

# ----------------------------------------

class AboutBox (Toplevel):

    def __init__(self, parent):
        Toplevel.__init__(self, parent)
        self.title('About TourEditor')
        self.transient(parent)
        ttk.Label(self, text='TourEditor '+__version__).grid (row=0, column=0)
        ttk.Label(self, text=u'© 2010-2011 Frank Pählke <frank@kette-links.de>').grid (row=1, column=0)
        ttk.Label(self, borderwidth=2, relief=SUNKEN,
                  text="""TourEditor is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TourEditor is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.""").grid (row=2, column=0)
        ttk.Button(self, text='OK', command=self.destroy).grid(row=3, column=0)
        self.grab_set()
        self.wait_window()

# ----------------------------------------

class TourEditor (Tk):
    "main application window"

    def __init__ (self):
        Tk.__init__ (self)
        self.title ('Tour Editor')
        l = ttk.Label (text='Encoding:')
        l.pack (side=LEFT)
        c = ttk.Combobox(values=_encodings)
        c.set ('iso-8859-1')
        c.pack (side=LEFT)
        self.encodingBox = c
        b = ttk.Button (text='Open File', command=self.open)
        b.pack (side=LEFT)
        b = ttk.Button (text='About', command=self.about)
        b.pack (side=LEFT)
        b = ttk.Button (text='Exit', command=self.quit)
        b.pack (side=LEFT)

    def open (self):
        filename = tkFileDialog.askopenfilename (
            title='Open File', filetypes=_filetypes)
        print repr(filename)
        if filename != None and filename != '':
            encoding = self.encodingBox.get()
            TourView (TourFile.read (filename, encoding))

    def about (self):
        aboutBox = AboutBox(self)

# ----------------------------------------
